# -*- coding: utf-8 -*-
# @Time    : 2020/9/18 11:33
# @Author  : Hui Wang
# @Email   : hui.wang@ruc.edu.cn

"""
SASRec
################################################

Reference:
    Wang-Cheng Kang et al. "Self-Attentive Sequential Recommendation." in ICDM 2018.

Reference:
    https://github.com/kang205/SASRec

"""

import torch
from torch import nn

from recbole.model.abstract_recommender import SequentialRecommender
from recbole.model.layers import TransformerEncoder
from recbole.model.layers import MultiIntentEncoder
from recbole.model.loss import BPRLoss


class DuoRec(SequentialRecommender):
    r"""
    SASRec is the first sequential recommender based on self-attentive mechanism.

    NOTE:
        In the author's implementation, the Point-Wise Feed-Forward Network (PFFN) is implemented
        by CNN with 1x1 kernel. In this implementation, we follows the original BERT implementation
        using Fully Connected Layer to implement the PFFN.
    """

    def __init__(self, config, dataset):
        super(DuoRec, self).__init__(config, dataset)

        # load parameters info
        # 2
        self.n_layers = config['n_layers']
        # 2
        self.n_heads = config['n_heads']
        # 64
        self.hidden_size = config['hidden_size']  # same as embedding_size
        # 256
        self.inner_size = config['inner_size']  # the dimensionality in feed-forward layer
        # 0.5
        self.hidden_dropout_prob = config['hidden_dropout_prob']
        # 0.5
        self.attn_dropout_prob = config['attn_dropout_prob']
        self.hidden_act = config['hidden_act']
        self.layer_norm_eps = config['layer_norm_eps']

        # 0.1
        self.lmd = config['lmd']
        # 0.1
        self.lmd_sem = config['lmd_sem']

        self.initializer_range = config['initializer_range']
        self.loss_type = config['loss_type']

        # define layers and loss
        self.item_embedding = nn.Embedding(self.n_items, self.hidden_size, padding_idx=0)
        self.position_embedding = nn.Embedding(self.max_seq_length, self.hidden_size)
        # self.trm_encoder = TransformerEncoder(
        #     n_layers=self.n_layers,
        #     n_heads=self.n_heads,
        #     hidden_size=self.hidden_size,
        #     inner_size=self.inner_size,
        #     hidden_dropout_prob=self.hidden_dropout_prob,
        #     attn_dropout_prob=self.attn_dropout_prob,
        #     hidden_act=self.hidden_act,
        #     layer_norm_eps=self.layer_norm_eps
        # )
        self.trm_encoder = MultiIntentEncoder(n_layers=self.n_layers,
                                              n_heads=self.n_heads,
                                              k_interests=config['k_interests'],
                                              hidden_size=self.hidden_size,
                                              seq_len=self.max_seq_length,
                                              inner_size=self.inner_size,
                                              hidden_dropout_prob=self.hidden_dropout_prob,
                                              attn_dropout_prob=self.attn_dropout_prob,
                                              hidden_act=self.hidden_act,
                                              layer_norm_eps=self.layer_norm_eps)

        self.LayerNorm = nn.LayerNorm(self.hidden_size, eps=self.layer_norm_eps)
        self.dropout = nn.Dropout(self.hidden_dropout_prob)

        if self.loss_type == 'BPR':
            self.loss_fct = BPRLoss()
        elif self.loss_type == 'CE':
            self.loss_fct = nn.CrossEntropyLoss()
        else:
            raise NotImplementedError("Make sure 'loss_type' in ['BPR', 'CE']!")

        self.ssl = config['contrast']
        self.tau = config['tau']
        self.sim = config['sim']
        self.batch_size = config['train_batch_size']
        self.mask_default = self.mask_correlated_samples(batch_size=self.batch_size)
        self.aug_nce_fct = nn.CrossEntropyLoss()
        self.sem_aug_nce_fct = nn.CrossEntropyLoss()

        # parameters initialization
        self.apply(self._init_weights)

    def _init_weights(self, module):
        """ Initialize the weights """
        if isinstance(module, (nn.Linear, nn.Embedding)):
            # Slightly different from the TF version which uses truncated_normal for initialization
            # cf https://github.com/pytorch/pytorch/pull/5617
            module.weight.data.normal_(mean=0.0, std=self.initializer_range)
            # module.weight.data = self.truncated_normal_(tensor=module.weight.data, mean=0, std=self.initializer_range)
        elif isinstance(module, nn.LayerNorm):
            module.bias.data.zero_()
            module.weight.data.fill_(1.0)
        if isinstance(module, nn.Linear) and module.bias is not None:
            module.bias.data.zero_()

    def truncated_normal_(self, tensor, mean=0, std=0.09):
        with torch.no_grad():
            size = tensor.shape
            tmp = tensor.new_empty(size + (4,)).normal_()
            valid = (tmp < 2) & (tmp > -2)
            ind = valid.max(-1, keepdim=True)[1]
            tensor.data.copy_(tmp.gather(-1, ind).squeeze(-1))
            tensor.data.mul_(std).add_(mean)
            return tensor

    def get_attention_mask(self, item_seq):
        """Generate left-to-right uni-directional attention mask for multi-head attention."""
        attention_mask = (item_seq > 0).long()
        extended_attention_mask = attention_mask.unsqueeze(1).unsqueeze(2)  # torch.int64
        # mask for left-to-right unidirectional
        max_len = attention_mask.size(-1)
        attn_shape = (1, max_len, max_len)
        subsequent_mask = torch.triu(torch.ones(attn_shape), diagonal=1)  # torch.uint8
        subsequent_mask = (subsequent_mask == 0).unsqueeze(1)
        subsequent_mask = subsequent_mask.long().to(item_seq.device)

        extended_attention_mask = extended_attention_mask * subsequent_mask
        extended_attention_mask = extended_attention_mask.to(dtype=next(self.parameters()).dtype)  # fp16 compatibility
        extended_attention_mask = (1.0 - extended_attention_mask) * -10000.0
        return extended_attention_mask

    def get_bi_attention_mask(self, item_seq):
        """Generate bidirectional attention mask for multi-head attention."""
        attention_mask = (item_seq > 0).long()
        extended_attention_mask = attention_mask.unsqueeze(1).unsqueeze(2)  # torch.int64
        # bidirectional mask
        extended_attention_mask = extended_attention_mask.to(dtype=next(self.parameters()).dtype)  # fp16 compatibility
        extended_attention_mask = (1.0 - extended_attention_mask) * -10000.0
        return extended_attention_mask

    # torch.Size([256, 50])  item_seq_len: tensor(256) [5,3,4,4,1,22,2,1,17,...]
    def forward(self, item_seq, item_seq_len):
        position_ids = torch.arange(item_seq.size(1), dtype=torch.long, device=item_seq.device)
        # torch.Size([256, 50])
        # position_ids = position_ids.unsqueeze(0).expand_as(item_seq)
        # torch.Size([256, 50, 64])
        position_embedding = self.position_embedding(position_ids)

        # torch.Size([256, 50, 64])
        item_emb = self.item_embedding(item_seq)
        # input_emb = item_emb + position_embedding
        input_emb = item_emb
        input_emb = self.LayerNorm(input_emb)
        # torch.Size([256, 50, 64])
        input_emb = self.dropout(input_emb)

        # torch.Size([256, 1, 50, 50])
        extended_attention_mask = self.get_attention_mask(item_seq)
        # extended_attention_mask = self.get_bi_attention_mask(item_seq)

        # list(2)   trm_output[0、1]: torch.Size([256, 50, 64])
        # trm_output = self.trm_encoder(input_emb, extended_attention_mask, output_all_encoded_layers=True)
        trm_output = self.trm_encoder(input_emb, position_embedding, output_all_encoded_layers=True)
        # torch.Size([256, 50, 64])
        output = trm_output[-1]
        # torch.Size([256, 64])
        output = self.gather_indexes(output, item_seq_len - 1)
        return output  # [B H]

    def calculate_loss(self, interaction):
        # torch.Size([256, 50])  self.ITEM_SEQ: 'item_id_list'
        item_seq = interaction[self.ITEM_SEQ]
        # Tensor(50)
        item_seq_len = interaction[self.ITEM_SEQ_LEN]
        # torch.Size([256, 64])  输出就是transformer的结果
        seq_output = self.forward(item_seq, item_seq_len)
        # Tensor(256) [113, 7306, 130, 4221, 2701, ...] self.POS_ITEM_ID: 'item_id'
        pos_items = interaction[self.POS_ITEM_ID]
        # self.loss_type：'CE'
        if self.loss_type == 'BPR':
            neg_items = interaction[self.NEG_ITEM_ID]
            pos_items_emb = self.item_embedding(pos_items)
            neg_items_emb = self.item_embedding(neg_items)
            pos_score = torch.sum(seq_output * pos_items_emb, dim=-1)  # [B]
            neg_score = torch.sum(seq_output * neg_items_emb, dim=-1)  # [B]
            loss = self.loss_fct(pos_score, neg_score)
        else:  # self.loss_type = 'CE'
            # torch.Size([12102, 64])
            test_item_emb = self.item_embedding.weight
            # torch.Size([256, 12102])
            logits = torch.matmul(seq_output, test_item_emb.transpose(0, 1))
            # tensor(9.4013, grad_fn=<NllLossBackward>)   loss_fct就是nn.CrossEntropyLoss
            loss = self.loss_fct(logits, pos_items)

        # Unsupervised NCE
        # self.ssl: 'us_x'
        if self.ssl in ['us', 'un']:
            aug_seq_output = self.forward(item_seq, item_seq_len)
            nce_logits, nce_labels = self.info_nce(seq_output, aug_seq_output, temp=self.tau,
                                                   batch_size=item_seq_len.shape[0], sim=self.sim)

            # nce_logits = torch.mm(seq_output, aug_seq_output.T)
            # nce_labels = torch.tensor(list(range(nce_logits.shape[0])), dtype=torch.long, device=item_seq.device)

            # if self.ssl == 'un':
            #     with torch.no_grad():
            #         alignment, uniformity = self.decompose(seq_output, aug_seq_output, seq_output,
            #                                                batch_size=item_seq_len.shape[0])

            loss += self.lmd * self.aug_nce_fct(nce_logits, nce_labels)

        # Supervised NCE
        if self.ssl in ['us', 'su']:
            sem_aug, sem_aug_lengths = interaction['sem_aug'], interaction['sem_aug_lengths']
            sem_aug_seq_output = self.forward(sem_aug, sem_aug_lengths)
            sem_nce_logits, sem_nce_labels = self.info_nce(seq_output, sem_aug_seq_output, temp=self.tau, batch_size=item_seq_len.shape[0], sim=self.sim)

            # sem_aug_2, sem_aug_lengths_2 = interaction['sem_aug_2'], interaction['sem_aug_lengths_2']
            # sem_aug_seq_output_2 = self.forward(sem_aug_2, sem_aug_lengths_2)
            # sem_nce_logits_2, sem_nce_labels_2 = self.info_nce(seq_output, sem_aug_seq_output_2, temp=self.tau, batch_size=item_seq_len.shape[0], sim=self.sim)

            # sem_nce_logits_3, sem_nce_labels_3 = self.info_nce(sem_aug_seq_output, sem_aug_seq_output_2, temp=self.tau, batch_size=item_seq_len.shape[0], sim=self.sim)

            # sem_nce_logits = torch.mm(seq_output, sem_aug_seq_output.T) / self.tau
            # sem_nce_labels = torch.tensor(list(range(sem_nce_logits.shape[0])), dtype=torch.long, device=item_seq.device)

            # with torch.no_grad():
            #     alignment, uniformity = self.decompose(seq_output, sem_aug_seq_output, seq_output,
            #                                            batch_size=item_seq_len.shape[0])

            # loss += self.lmd_sem * (self.aug_nce_fct(sem_nce_logits, sem_nce_labels) + self.aug_nce_fct(sem_nce_logits_2, sem_nce_labels_2))
            loss += (self.lmd_sem * self.aug_nce_fct(sem_nce_logits, sem_nce_labels))
        if self.ssl == 'sm':
            # sem_aug, sem_aug_lengths = interaction['sem_aug'], interaction['sem_aug_lengths']
            # sem_aug_seq_output = self.forward(sem_aug, sem_aug_lengths)
            # sem_nce_logits, sem_nce_labels = self.info_nce(seq_output, sem_aug_seq_output, temp=self.tau, batch_size=item_seq_len.shape[0], sim=self.sim)

            sem_aug_2, sem_aug_lengths_2 = interaction['sem_aug_2'], interaction['sem_aug_lengths_2']
            sem_aug_seq_output_2 = self.forward(sem_aug_2, sem_aug_lengths_2)
            sem_nce_logits_2, sem_nce_labels_2 = self.info_nce(seq_output, sem_aug_seq_output_2, temp=self.tau, batch_size=item_seq_len.shape[0], sim=self.sim)

            # sem_nce_logits_3, sem_nce_labels_3 = self.info_nce(sem_aug_seq_output, sem_aug_seq_output_2, temp=self.tau, batch_size=item_seq_len.shape[0], sim=self.sim)

            # sem_nce_logits = torch.mm(seq_output, sem_aug_seq_output.T) / self.tau
            # sem_nce_labels = torch.tensor(list(range(sem_nce_logits.shape[0])), dtype=torch.long, device=item_seq.device)

            # with torch.no_grad():
            #     alignment, uniformity = self.decompose(seq_output, sem_aug_seq_output, seq_output,
            #                                            batch_size=item_seq_len.shape[0])

            # loss += self.lmd_sem * (self.aug_nce_fct(sem_nce_logits, sem_nce_labels) + self.aug_nce_fct(sem_nce_logits_2, sem_nce_labels_2))
            loss += (self.lmd_sem * self.aug_nce_fct(sem_nce_logits_2, sem_nce_labels_2))

        if self.ssl == 'sc':
            sem_aug_3, sem_aug_lengths_3 = interaction['sem_aug_3'], interaction['sem_aug_lengths_3']
            sem_aug_seq_output_3 = self.forward(sem_aug_3, sem_aug_lengths_3)
            sem_nce_logits_3, sem_nce_labels_3 = self.info_nce(seq_output, sem_aug_seq_output_3, temp=self.tau, batch_size=item_seq_len.shape[0], sim=self.sim)
            loss += (self.lmd_sem * self.aug_nce_fct(sem_nce_logits_3, sem_nce_labels_3))

        if self.ssl == 'us_x':
            # torch.Size([256, 64])   可以想象每次forward的dropout中被隐藏的单元都是变化的
            aug_seq_output = self.forward(item_seq, item_seq_len)

            # torch.Size([256, 50])   Tensor(256)   可以全局搜sem_aug，可发现在sequential_dataloader.py里
            sem_aug, sem_aug_lengths = interaction['sem_aug'], interaction['sem_aug_lengths']
            # torch.Size([256, 64])
            sem_aug_seq_output = self.forward(sem_aug, sem_aug_lengths)

            # torch.Size([512, 511])    Tensor(512)
            sem_nce_logits, sem_nce_labels = self.info_nce(aug_seq_output, sem_aug_seq_output, temp=self.tau,
                                                           batch_size=item_seq_len.shape[0], sim=self.sim)

            # tensor(12.4259, grad_fn=<AddBackward0>)  论文式15
            loss += self.lmd_sem * self.aug_nce_fct(sem_nce_logits, sem_nce_labels)

            # with torch.no_grad():
            #     alignment, uniformity = self.decompose(aug_seq_output, sem_aug_seq_output, seq_output,
            #                                            batch_size=item_seq_len.shape[0])

        return loss

    # 256
    def mask_correlated_samples(self, batch_size):
        # 512
        N = 2 * batch_size
        # torch.Size([512, 512])
        mask = torch.ones((N, N), dtype=bool)
        # 主对角线为0
        mask = mask.fill_diagonal_(0)
        for i in range(batch_size):
            # 第二分块的对角线为0
            mask[i, batch_size + i] = 0
            # 第三分块的对角线为0，因为对称
            mask[batch_size + i, i] = 0
        return mask

    # torch.Size([256, 64])   torch.Size([256, 64])  temp: 1  sim: 'dot'
    def info_nce(self, z_i, z_j, temp, batch_size, sim='dot'):
        """
        We do not sample negative examples explicitly.
        Instead, given a positive pair, similar to (Chen et al., 2017), we treat the other 2(N − 1) augmented examples within a minibatch as negative examples.
        """
        N = 2 * batch_size

        # torch.Size([512, 64])
        z = torch.cat((z_i, z_j), dim=0)

        if sim == 'cos':
            sim = nn.functional.cosine_similarity(z.unsqueeze(1), z.unsqueeze(0), dim=2) / temp
        elif sim == 'dot':
            # torch.Size([512, 512])
            sim = torch.mm(z, z.T) / temp

        # Tensor(256)  主轴往上数256次的那一斜行  这样能避开与自己群体的那部分值
        sim_i_j = torch.diag(sim, batch_size)
        # Tensor(256)   sim_i_j和sim_j_i值是相同的，因为A和a的点击与a和A的点击相同，但是与不同的负样本拼接后就会不一样，所以仍有一定存在价值
        sim_j_i = torch.diag(sim, -batch_size)

        # torch.Size([512, 1])
        positive_samples = torch.cat((sim_i_j, sim_j_i), dim=0).reshape(N, 1)
        if batch_size != self.batch_size:
            mask = self.mask_correlated_samples(batch_size)
        else:
            # torch.Size([512, 512])    todo
            mask = self.mask_default
        # torch.Size([512, 510])  mask里面每一行有两个0，分别在主对角线和第二第三分块对角线，所以，每一行少了两个值，变成了510，第二第三分块去掉式因为那里位置是正样本值
        negative_samples = sim[mask].reshape(N, -1)

        # Tensor(512)
        labels = torch.zeros(N).to(positive_samples.device).long()
        # torch.Size([512, 511])   上面的labels全为0是因为这里的正样本位置在首位
        logits = torch.cat((positive_samples, negative_samples), dim=1)
        return logits, labels

    def decompose(self, z_i, z_j, origin_z, batch_size):
        """
        We do not sample negative examples explicitly.
        Instead, given a positive pair, similar to (Chen et al., 2017), we treat the other 2(N − 1) augmented examples within a minibatch as negative examples.
        """
        N = 2 * batch_size

        z = torch.cat((z_i, z_j), dim=0)

        # pairwise l2 distace
        sim = torch.cdist(z, z, p=2)

        sim_i_j = torch.diag(sim, batch_size)
        sim_j_i = torch.diag(sim, -batch_size)

        positive_samples = torch.cat((sim_i_j, sim_j_i), dim=0).reshape(N, 1)
        alignment = positive_samples.mean()

        # pairwise l2 distace
        sim = torch.cdist(origin_z, origin_z, p=2)
        mask = torch.ones((batch_size, batch_size), dtype=bool)
        mask = mask.fill_diagonal_(0)
        negative_samples = sim[mask].reshape(batch_size, -1)
        uniformity = torch.log(torch.exp(-2 * negative_samples).mean())

        return alignment, uniformity

    def predict(self, interaction):
        item_seq = interaction[self.ITEM_SEQ]
        item_seq_len = interaction[self.ITEM_SEQ_LEN]
        test_item = interaction[self.ITEM_ID]
        seq_output = self.forward(item_seq, item_seq_len)
        test_item_emb = self.item_embedding(test_item)
        scores = torch.mul(seq_output, test_item_emb).sum(dim=1)  # [B]
        return scores

    def full_sort_predict(self, interaction):
        item_seq = interaction[self.ITEM_SEQ]
        item_seq_len = interaction[self.ITEM_SEQ_LEN]
        seq_output = self.forward(item_seq, item_seq_len)
        test_items_emb = self.item_embedding.weight
        scores = torch.matmul(seq_output, test_items_emb.transpose(0, 1))  # [B n_items]
        return scores
